/*******************************************************************************
** INTEL CONFIDENTIAL
**
** Copyright (c) 2020-2022 Intel Corporation All Rights Reserved
**
** This software and the related documents are Intel copyrighted materials,
** and your use of them is governed by the express license under which they
** were provided to you ("License"). Unless the License provides otherwise,
** you may not use, modify, copy, publish, distribute, disclose or transmit
** this software or the related documents without Intel's prior written
** permission.
**
** The source code contained or described herein and all documents related to
** the source code ("Material") are owned by Intel Corporation or its suppliers
** or licensors.
**
** This software and the related documents are provided as is,
** with no express or implied warranties, other than those that are expressly
** stated in the License.
**
*******************************************************************************/

using Ipf;
using Ipf.CLI;
using Newtonsoft.Json.Linq;
using System;
using System.Runtime.InteropServices;  // For GCHandle

namespace ConsoleApp1
{
    internal class Program
    {

        // An object that can be passed to RegisterEvent as context.
        // A pointer to this object will be passed into the SampleEventCallback
        public class ContextObject
        {
            public void Hello(String event_data)
            {
                Console.WriteLine("Hello from ContextObject! Event data is " + event_data);
            }
        }

        // The EventCallbackDelegate that is passed into RegisterEvent and UnregisterEvent
        public static void SampleEventCallback(String event_name, String event_data, System.IntPtr context)
        {
            Console.WriteLine("Received event from " + event_name + ". Event data: " + event_data);

            // Convert the context pointer into a GCHandle, then convert the handle into our object
            GCHandle handle = GCHandle.FromIntPtr(context);
            ContextObject context_object = (handle.Target as ContextObject);

            // Call a function on our context object
            context_object.Hello(event_data);
        }

        // Create an instance of the callback delegate so that
        // it doesn't get garbage collected
        static EventCallbackDelegate callbackInstance = SampleEventCallback;

        private static void ConsolePrint(String field, String value)
        {
            Console.WriteLine(field);
            Console.WriteLine(value);
            Console.WriteLine();
        }

        private static void Main(string[] args)
        {
            String config = "";
            try
            {
                if (RuntimeInformation.IsOSPlatform(OSPlatform.Windows))
                {
                    config = @"
                    {
                        ""schema_type"": ""provider_config"",
                        ""schema_revision"": 1,
                        ""body"" : {
                            ""local_providers"": [
                                    ""SampleCameraProvider.dll"",
                                    ""SampleBatteryProvider.dll""
                                 ],
							""init"": {
								""Company.Org.Battery"":
									{
										""manufacturer"": ""UpdatedBatteryManufacturer"",
										""model"": ""MUO6""
									}
								}
							}
                    }
                    ";
                }
                else if (RuntimeInformation.IsOSPlatform(OSPlatform.Linux))
                {
                    config = @"
                    {
                        ""schema_type"": ""provider_config"",
                        ""schema_revision"": 1,
                        ""body"" : {
                            ""local_providers"": [
                                    ""libSampleCameraProvider.so"",
                                    ""libSampleBatteryProvider.so""
                                 ],
							""init"": {
                                ""Company.Org.Battery"":
									{
                                    ""manufacturer"": ""UpdatedBatteryManufacturer"",
									""model"": ""MUO6""

                                    }
                            }
                        }
                    }
                    ";
                }

                JObject provider_config = JObject.Parse(config);

                // Initialize the ClientApi with the JSON config
                ClientApiJson ipf = new ClientApiJson(provider_config);

                // Get ClientApi Properties
                JToken node = ipf.GetProperties();
                ConsolePrint("ipf.GetProperties()", node.ToString());

                // Get all objects and properties under Platform
                node = ipf.GetNode("Platform");
                ConsolePrint("ipf.GetNode(\"Platform\")", node.ToString());

                // Get all objects and properties under Camera Provider
                node = ipf.GetNode("Platform.Extension.Camera");
                ConsolePrint("ipf.GetNode(\"Platform.Extension.Camera\")", node.ToString());

                // Get Data Schema for Camera Provider
                JObject schema = ipf.GetSchema("Platform.Extension.Camera", SchemaType.DATA);
                ConsolePrint("ipf.GetSchema(\"Platform.Extension.Camera\")", schema.ToString());

                // Query node to get objects and properties at any level in the node tree
                JArray queryResult = ipf.QueryNode("Platform.Extension.Camera");
                ConsolePrint("ipf.QueryNode(\"Platform.Extension.Camera\")", queryResult.ToString());

                // Set fps property of Camera
                ipf.SetValue("Platform.Extension.Camera.fps", 100);
                JToken updatedFps = ipf.GetValue("Platform.Extension.Camera.fps");
                ConsolePrint("ipf.GetValue(\"Platform.Extension.Camera.fps\")", updatedFps.ToString());

                // Get init parameter information from Battery provider
                node = ipf.GetNode("Platform.Extension.Battery");
                ConsolePrint("ipf.GetNode(\"Platform.Extension.Battery\")", node.ToString());

                // Get manufacturer information which is a init parameter
                node = ipf.GetNode("Platform.Extension.Battery.manufacturer");
                ConsolePrint("ipf.GetNode(\"Platform.Extension.Battery.manufacturer\")", node.ToString());

                // Set state (enum) property of Battery
                ipf.SetValue("Platform.Extension.Battery.state", "charged");

                try
                {
                    // Set state (enum) property of Battery
                    ipf.SetValue("Platform.Extension.Battery.is_ion_battery", true);
                }
                catch (IpfManagedException.ProviderException)
                {
                    Console.WriteLine("Set value is not supported");
                }

                // Get the "state" property from the Battery
                JToken updatedState = ipf.GetValue("Platform.Extension.Battery.state");
                ConsolePrint("ipf.GetValue(\"Platform.Extension.Battery.state\")", updatedState.ToString());

                // Get child property value under Battery Provider
                JToken childState = ipf.GetValue("Platform.Extension.Battery.childCell.state");
                ConsolePrint("ipf.GetValue(\"Platform.Extension.Battery.childCell.state\")", childState.ToString());


                // Create a context object and get a handle to it from the
                // garbage collector.
                // 
                // Calling Alloc() prevents the object from being garbage collected.
                //
                // We must free the handle after unregistering the event or the
                // object will be leaked.
                ContextObject context_object = new ContextObject();
                GCHandle context_handle = GCHandle.Alloc(context_object);

                // Register for an event.  Passing a context is optional
                ipf.RegisterEvent("Platform.Extension.Battery.batteryFullyCharged",
                  callbackInstance, GCHandle.ToIntPtr(context_handle));

                // Execute a command on the battery provider
                string cmd = @"
                     {
                         ""commandName"": ""value1""
                     }
                 ";
                JObject cmdobj = JObject.Parse(cmd);
                JObject result = ipf.ExecuteCommand("Platform.Extension.Battery", cmdobj);
                ConsolePrint("ipf.ExecuteCommand(\"Platform.Extension.Battery, cmdobj\")", result?.ToString());

                // Unregister for the event
                ipf.UnregisterEvent("Platform.Extension.Battery.batteryFullyCharged",
                callbackInstance);

                string cmd_simulate_plug = @"
					{
						""simulatePlug"" : {
							 ""instance_name"": ""ChildBatteryProvider"",
							 ""manufacturer"" : ""Dell"",
							 ""model"" : ""D6OU""
						}
					}
					";

                string cmd_simulate_unplug = @"
					{
						""simulateUnplug"" :{ 
							""instance_name"": ""ChildBatteryProvider""
						}
					}
					";

                // Execute command creates a child provider with "ChildBatteryProvider" instance name
                JObject cmd_sim_plug = JObject.Parse(cmd_simulate_plug);
                result = ipf.ExecuteCommand("Platform.Extension.Battery", cmd_sim_plug);
                ConsolePrint("ipf.ExecuteCommand(\"Platform.Extension.Battery, cmd_sim_plug\")", result?.ToString());

                // Call child provider's execute command
                result = ipf.ExecuteCommand("Platform.Extension.Battery.ChildBatteryProvider", cmd_sim_plug);
                ConsolePrint("ipf.ExecuteCommand(\"Platform.Extension.Battery.ChildBatteryProvider, cmd_sim_plug\")", result?.ToString());

                JObject cmd_sim_unplug = JObject.Parse(cmd_simulate_unplug);
                result = ipf.ExecuteCommand("Platform.Extension.Battery", cmd_sim_unplug);
                ConsolePrint("ipf.ExecuteCommand(\"Platform.Extension.Battery, cmd_sim_unplug\")", result?.ToString());


                // Free the context handle ONLY after unregistering for the event.
                // If the event is raised after the context_handle is freed, it will
                // cause a System.NullReferenceException
                context_handle.Free();
            }

            // Catch IPF Exceptions
            catch (IpfManagedException.IpfException e)
            {
                Console.WriteLine("Got IpfException when accessing the namespace:");
                Console.WriteLine(e.Message);
            }
        }
    }
}
